#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "Nebula_Device_AV.h"
#include "playback.h"
#include <windows.h>


extern AV_Client gClientInfo[MAX_CLIENT_NUMBER];

static void usleep(long long usec) {
    HANDLE timer;
    LARGE_INTEGER interval;
    interval.QuadPart = -(10 * usec);

    timer = CreateWaitableTimer(NULL, TRUE, NULL);
    SetWaitableTimer(timer, &interval, 0, NULL, NULL, 0);
    WaitForSingleObject(timer, INFINITE);
    CloseHandle(timer);
}

static int GetClientPlaybackMode(int sid)
{
    AV_Client *p = &gClientInfo[sid];
    return p->enable_record_video;
}

static int IsVideoData(unsigned short codec_id)
{
    if(codec_id >= MEDIA_CODEC_VIDEO_MPEG4 && codec_id <= MEDIA_CODEC_VIDEO_HEVC)
        return 1;
    return 0;
}

static int IsAudioData(unsigned short codec_id)
{
    if(codec_id >= MEDIA_CODEC_AUDIO_AAC_RAW && codec_id <= MEDIA_CODEC_AUDIO_G726)
        return 1;
    return 0;
}

static int OpenRecordFile(AV_Client *info, int *fps)
{
    int file_fps = 0;
    char info_file[128] = {0};
    char record_file[128] = {0};
    char record_audio[128] = {0};
    snprintf(info_file, sizeof(info_file), "%s.info", info->playback_file);
    snprintf(record_file, sizeof(record_file), "%s.bin", info->playback_file);
    snprintf(record_audio, sizeof(record_audio), "%s.raw", info->playback_file);

    FILE *file_info_fd = fopen(info_file, "r");
    if(file_info_fd == NULL) {
        printf("Open file info %s fail!!\n", info_file);
        return -1;
    } else {
        char line_buf[128] = {0};
        //get FPS from info file
        char *ret_str = fgets(line_buf, sizeof(line_buf), file_info_fd);
        if (ret_str == NULL) {
            fclose(file_info_fd);
            return -1;
        }
        sscanf(line_buf, "FPS %d", &file_fps);
    }

    FILE *file_fd = fopen(record_file, "r");
    if(file_fd == NULL) {
        printf("Open file %s fail!!\n", record_file);
        fclose(file_info_fd);
        return -1;
    }

    FILE *audio_fd = fopen(record_audio, "r");
    if(audio_fd == NULL) {
        printf("Open file %s fail!!\n", record_audio);
        fclose(file_info_fd);
        fclose(file_fd);
        return -1;
    }

    *fps = file_fps;
    info->playback_file_info_fd = file_info_fd;
    info->playback_file_fd = file_fd;
    info->playback_audio_fd = audio_fd;
    return 0;
}

static int CloseRecordFile(AV_Client *info)
{
    if(info->playback_file_fd){
        fclose(info->playback_file_fd);
        info->playback_file_fd = NULL;
    }
    if(info->playback_file_info_fd){
        fclose(info->playback_file_info_fd);
        info->playback_file_info_fd = NULL;
    }
    if(info->playback_audio_fd){
        fclose(info->playback_audio_fd);
        info->playback_audio_fd = NULL;
    }
    memset(info->playback_file, 0, sizeof(info->playback_file));
    return 0;
}

static int GetFrameInfo(FILE *info_file_ptr, ENUM_FRAMEFLAG *frame_type, unsigned int *frame_size, unsigned int *last_frame)
{
	char line_buf[512] = {0};
	char type;
	unsigned int pos;

	char *ret = fgets(line_buf, 512, info_file_ptr);
	if (ret == NULL)
		return -1;

	sscanf(line_buf, "%c %d %d", &type, &pos, frame_size);
	*frame_type = type == 'I' ? IPC_FRAME_FLAG_IFRAME : IPC_FRAME_FLAG_PBFRAME;
    
    if( feof(info_file_ptr) ) {
        *last_frame = 1;
    } else {
        *last_frame = 0;
    }
	return 0;
}

static int GetRecordFileData(const AV_Client *info, char *buffer, unsigned int buf_size, FRAMEINFO_t *frame_info, int video_sleep_ms, int audio_sleep_ms)
{
    int ret = 0;
    static int last_video_time = 0, last_audio_time = 0;
    ENUM_FRAMEFLAG frame_type;
	unsigned int frame_size = 0, last_frame = 0;

    // get video frame
    if((GetTimeStampMs() - last_video_time) >= video_sleep_ms){
        ret = GetFrameInfo(info->playback_file_info_fd, &frame_type, &frame_size, &last_frame);
        if(ret == 0){
            ret = fread(buffer, 1, frame_size, info->playback_file_fd);
            if(ret <= 0) {
                // End of file or file broken
                printf("End of record file\n");
                return -1;
            }
            frame_info->codec_id = MEDIA_CODEC_VIDEO_H264;
            frame_info->flags = frame_type;
            frame_info->timestamp = GetTimeStampMs();
            last_video_time = frame_info->timestamp;
            frame_info->onlineNum = 1;
            if(last_frame == 1){
                //The last frame in record file
                printf("last frame\n");
                frame_info->tags = 1;
            }
        } else {
            // End of file or file broken
            printf("End of record file\n");
            frame_size = -1;
        }

        return frame_size;
    }

    // get audio frame
    if((GetTimeStampMs() - last_audio_time + 1) >= audio_sleep_ms){
        ret = fread(buffer, 1, AUDIO_FRAME_SIZE, info->playback_audio_fd);
        if(ret == 0){
            printf("last audio frame, rewind playback audio\n");
			rewind(info->playback_audio_fd);
            return 0;
        } else if(ret < 0){
            printf("audio file error\n");
            return -1;
        }

        //memset(frame_info, 0, sizeof(FRAMEINFO_t));
        frame_info->codec_id = AUDIO_CODEC;
        frame_info->flags = (AUDIO_SAMPLE_8K << 2) | (AUDIO_DATABITS_16 << 1) | AUDIO_CHANNEL_MONO;
        frame_info->timestamp = GetTimeStampMs();
        last_audio_time = frame_info->timestamp;

        return ret;
    }

    return 0;
}

//########################################################
//#Thread - Send record file
//########################################################
static void WINAPI ThreadRecordFileData(void *arg)
{
    int sid = *((int *)arg);
    free(arg);
    printf("%s start SID[%d]\n", __func__, sid);

    AV_Client *info = &gClientInfo[sid];
    int av_index = StartAvServer(sid, AV_PLAYBACK_CHANNEL, 30, 1024, NULL);
    if (av_index < 0) {
        printf("StartAvServer error!! avIndex[%d]\n", av_index);
        goto THREAD_RECORD_EXIT;
    }
    int data_size = 0, file_fps = 0, video_sleep_ms = 0, audio_sleep_ms = 0, ret = 0;
    char data_buf[VIDEO_BUF_SIZE] ={0};
    int cache_data = 0;
    int end_of_file_or_file_broken = 0;
    FRAMEINFO_t frame_info;

    RegEditClientPlaybackMode(sid, PLAYBACK_START);
    ret = OpenRecordFile(info, &file_fps);
    if(ret != 0) {
        end_of_file_or_file_broken = 1;
        printf("Can't open record file correctly\n");
    }
    video_sleep_ms = (1000/file_fps)-1;
    audio_sleep_ms = (1000/AUDIO_FPS)-1;

    while(GetClientPlaybackMode(sid) != PLAYBACK_STOP) {
        if(GetClientPlaybackMode(sid) == PLAYBACK_PAUSE || 1 == end_of_file_or_file_broken) {
            usleep(500*1000);
            continue;
        }
        if(cache_data == 0) {
            memset(&frame_info, 0, sizeof(FRAMEINFO_t));
            data_size = GetRecordFileData(info, data_buf, sizeof(data_buf), &frame_info, video_sleep_ms, audio_sleep_ms);
            if(data_size < 0){
                end_of_file_or_file_broken = 1;
                continue;
            } else if(data_size == 0){
                usleep(1000);
                continue;
            }
        }

        if(IsVideoData(frame_info.codec_id)){
            ret = avSendFrameData(av_index, data_buf, data_size, (void *)&frame_info, sizeof(FRAMEINFO_t));
        } else if(IsAudioData(frame_info.codec_id)){
            ret = avSendAudioData(av_index, data_buf, data_size, (void *)&frame_info, sizeof(FRAMEINFO_t));
        } else {
            printf("Unknown data codec[%d]\n", frame_info.codec_id);
            end_of_file_or_file_broken = 1;
            continue;
        }

        if (ret == AV_ER_EXCEED_MAX_SIZE) {
            // AV resend buffer is full. 
            // Need to keep and re-send this frame until avSendFrameData or avSendAudioData retrun AV_ER_NoERROR
            cache_data = 1;
            usleep(5000);
            continue;
        } else if (ret < 0) {
            printf("[SendRecordData] avIndex[%d] error[%d]\n", av_index, ret);
            break;
        } else {
            cache_data = 0;
            printf("[SendRecordData] avIndex[%d] frameSize[%d]\n", av_index, data_size);
        }
    }

THREAD_RECORD_EXIT:
    WaitForSingleObject(info->lock, INFINITE);
    CloseRecordFile(info);
    RegEditClientPlaybackMode(sid, PLAYBACK_STOP);
    ReleaseMutex(info->lock);

    if(av_index >= 0) {
        avServStop(av_index);
        printf("[%s:%d] avServStop playback avIndex[%d]\n", __func__, __LINE__, av_index);
    }
    printf("%s: exit!! SID[%d]\n", __func__, sid);
    return;
}

//########################################################
//# Handle client command for record file
//########################################################
int HandlePlaybackControl(int sid, int ctrl_value, const char* file_name)
{
    AV_Client *p = &gClientInfo[sid];
    int ret = 0;

    WaitForSingleObject(p->lock, INFINITE);
    if(ctrl_value == PLAYBACK_START) {
        if (GetClientPlaybackMode(sid) == PLAYBACK_START) {
            ReleaseMutex(p->lock);
            printf("Playback already started!!\n");
            return -1;
        } else if (GetClientPlaybackMode(sid) == PLAYBACK_PAUSE) {
            RegEditClientPlaybackMode(sid, PLAYBACK_START);
            return 0;
        }
        // check if the record file exist or not
        char file_path[128] = "";
        snprintf(file_path, sizeof(file_path), "./video/%s.bin", file_name);
        FILE *fp = fopen(file_path, "rb");
        if(fp == NULL) {
            printf("Open file %s fail!!\n", file_path);
            ReleaseMutex(p->lock);
            return -1;
        }
        fclose(fp);
        snprintf(p->playback_file, sizeof(p->playback_file), "./video/%s", file_name);

        HANDLE thread_id = 0;
        int *tmp_sid = (int *)malloc(sizeof(int));
        *tmp_sid = sid;

        thread_id = CreateThread(NULL, 0, ThreadRecordFileData, (void *)tmp_sid, 0, NULL);
        if (thread_id == 0) {
            printf("%s() pthread_create ThreadRecordFileData ret[%d]\n", __func__, ret);
            free(tmp_sid);
            ReleaseMutex(p->lock, INFINITE);
            return -1;
        } else {
            CloseHandle(thread_id);
        }

    } else if(ctrl_value == PLAYBACK_STOP) {
        RegEditClientPlaybackMode(sid, PLAYBACK_STOP);
    } else if(ctrl_value == PLAYBACK_PAUSE) {
        RegEditClientPlaybackMode(sid, PLAYBACK_PAUSE);
    }
    ReleaseMutex(p->lock);

    return 0;
}

void GetRecordFileList(unsigned int start_time, unsigned int end_time, const char *event_type, TUTKJsonObject **response)
{
    // In this sample we ignore the search criteria and response the hard coded file list.
    char json[] = "{\"value\": [{\"eventType\": \"motion\",\"timestamp\": 1559754000,\"duration\": 300,\"fileName\": \"frames\", \"fileSize\": 6661927}]}";
    TUTK_Json_Obj_Create_From_String(json, response);
}

void RegEditClientPlaybackMode(int sid, int playback_mode)
{
    AV_Client *p = &gClientInfo[sid];
    p->enable_record_video = playback_mode;
}


